/*
  x_debug - a tiny debugger using DWARF2 XML
  Copyright (C) 2002, by gondow@jaist.ac.jp, All rights reserved.
  $Id: x_debug.c,v 1.7 2004/01/19 07:24:42 gondow Exp gondow $
 */
/* ======================================================================== */
#include <stdio.h>             /* for printf, fprintf, sscanf */
#include <stdlib.h>            /* for malloc, free, strtol */
#include <sys/types.h>         /* for open, lseek, fork, kill, setpgid */
#include <sys/stat.h>          /* for open, fstat */
#include <fcntl.h>             /* for open */
#include <unistd.h>            /* for lseek, write, read, fork, execve, 
				  getpid, setpgid, unlink */
#include <assert.h>            /* for assert */
#include <procfs.h>            /* for proc file system (proc(4)) */
#include <sys/fault.h>         /* for FLTBPT, FLTTRACE, FLTBOUNDS, FLTACCESS,
				  FLTWATCH */
#include <stdarg.h>            /* for va_start, va_arg, va_end */
#include <signal.h>            /* for sigaction, kill */
#include <sys/wait.h>          /* for waitpid */
#include <ctype.h>             /* for isspace, isnumber */
#include <string.h>            /* for strlen, strncmp, strdup */
#include <readline/readline.h> /* for GNU readline */
#include <readline/history.h>  /* for GNU readline */
#include <siginfo.h>           /* for siginfo (see siginfo(3HEAD)) */
#include <sys/reg.h>           /* for struct rwindow */

#include <sys/mman.h>          /* for mmap */
#include <sys/time.h>	       /* for gettimeofday */

/* for libxml2 library */
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/tree.h>
#include <libxml/debugXML.h>

#define BUFLEN   1024          /* for lineno2addr(), local2addr() */
/* ======================================================================== */
struct proc_fd {
    int     ctl;    /* file descriptor for /proc/<pid>/ctl */
    int     status; /* ditto           for /proc/<pid>/status */
    int     as;     /* ditto           for /proc/<pid>/as (address space) */
};

enum command_type {
    CMD_NONE,  CMD_RUN,   CMD_QUIT,  CMD_CONT, CMD_STEPI, CMD_STEP,
    CMD_BREAK, CMD_PRINT, CMD_WHERE, CMD_WATCH,
    N_CMD        /* this must be the last here */
};

/* used in try_wait() */
enum {
    WAIT_FAILURE, WAIT_SUCCESS
};

struct command {
    enum command_type type;
    char *line;
    char *arg;
    int  *addr;
    int  size;
    char fmt;
};

struct break_point {
    int  *addr;
    int   data;
};

struct location {
    char *dir;
    char *file;
    char *func;
    unsigned int line;
};
/* ======================================================================== */
static int      e_open   (const char *path, int oflag);
static ssize_t  e_write  (int fd, const void *buf, size_t nbyte);
static ssize_t  e_read   (int fd, void *buf, size_t nbyte);
static off_t    e_lseek  (int fd, off_t offset, int whence);
static offset_t e_llseek (int fd, offset_t offset, int whence);
static void    *e_malloc (size_t size);
static void     open_proc  (pid_t pid);
static void     close_proc (void);

static void send_ctl_pcstop   (long cmd);
static void send_ctl_pcrun    (long data);
static void send_ctl_pcsexit  (int sysnum, ...);
static void send_ctl_pcsfault (int fltnum, ...);

static void get_pstatus (void);
static int  try_wait (void);

static void init          (struct command *cmd_p, int argc, char *argv []);
static void read_command  (struct command *cmd_p);
static void parse_command (struct command *cmd_p);
static void dump_command  (struct command *cmd_p);

static void do_run   (struct command *cmd_p, char *aout);
static void do_quit  (struct command *cmd_p);
static void do_cont  (struct command *cmd_p);
static void do_stepi (struct command *cmd_p, int print_status);
static void do_step  (struct command *cmd_p);
static void do_break (struct command *cmd_p);
static void do_print (struct command *cmd_p);
static void do_where (struct command *cmd_p);
static void do_watch (struct command *cmd_p);

static void kill_child (void);
static void wait_child (int print_status);
static int  adjust_if_breakpoint (void);
static void *get_pc (void);
static void *get_sp (void);
static void *get_fp (void);
static void get_kwsa (void *kwsa, struct rwindow *rwindow);
static char **make_argv (char *aout, char *line);
static void sigint_handler (int signo);
static void sigchld_handler (int signo);
static void print_child_state (void);

static xmlNodeSetPtr eval_xpath (char *xpath, xmlNodePtr from);
static int *sym2addr (char *sym);
static void addr2lineno (int *addr);
static void show_line (void);

static int *lineno2addr (int line);
static int *local2addr (char *sym);
/* ======================================================================== */
static struct proc_fd  proc_fd;
static struct pstatus  pstatus;
static struct location location;
static pid_t           pid = 0;
static char *cmd_name [N_CMD];
static struct break_point break_point = {NULL, 0};
static unsigned int trap_ta1 = 0x91d02001; /* machine code for "ta 1" */
static char *sig_name [] = { NULL /*0*/,
    "HUP"  /* 1*/, "INT"  /* 2*/, "QUIT" /* 3*/, "ILL"  /* 4*/, "TRAP" /* 5*/,
    "IOT"  /* 6*/, "EMT"  /* 7*/, "FPE"  /* 8*/, "KILL" /* 9*/, "BUS"  /*10*/,
    "SEGV" /*11*/, "SYS"  /*12*/, "PIPE" /*13*/, "ALRM" /*14*/, "TERM" /*15*/,
    "USR1" /*16*/, "USR2" /*17*/, "CHLD" /*18*/, "PWR"  /*19*/, "WINCH" /*20*/,
    "URG"  /*21*/, "POLL" /*22*/, "STOP" /*23*/, "TSTP" /*24*/, "CONT" /*25*/,
    "TTIN" /*26*/, "TTOU" /*27*/, "VTALRM"/*28*/, "PROF"/*29*/, "XCPU" /*30*/,
    "XFSZ" /*31*/, "WAITING"/*32*/, "LWP"/*33*/, "FREEZE"/*34*/, "THAW"/*35*/,
    "CANCEL" /*36*/, "LOST" /*37*/, NULL
};

static xmlDocPtr  xml_doc  = NULL;
static xmlNodePtr xml_root = NULL;
/* ======================================================================== */
static int
e_open  (const char *path, int oflag)
{
    int fd;

    fd = open (path, oflag);
    if (fd < 0) {
	perror ("open");
	exit (1);
    }
    return fd;
}

static ssize_t
e_write (int fd, const void *buf, size_t nbyte)
{
    ssize_t result;

    result = write (fd, buf, nbyte);
    if (result < 0) {
	if (try_wait () != WAIT_SUCCESS) {
	    perror ("write");
	}
    }
    return result;
}

static ssize_t
e_read  (int fd, void *buf, size_t nbyte)
{
    ssize_t result;

    result = read (fd, buf, nbyte);
    if (result < 0) {
	if (try_wait () != WAIT_SUCCESS) {
	    perror ("read");
	}
    }
    return result;
}

static off_t
e_lseek (int fd, off_t offset, int whence)
{
    off_t result;

    result = lseek (fd, offset, whence);
    if (result < 0) {
	perror ("lseek"); exit (1);
    }
    return result;
}

static offset_t
e_llseek (int fd, offset_t offset, int whence)
{
    offset_t result;

    result = llseek (fd, offset, whence);
    if (result < 0) {
	perror ("llseek"); exit (1);
    }
    return result;
}

static void *
e_malloc (size_t size)
{ 
    void *p;

    p = malloc (size);
    if (p == NULL)
	perror ("malloc");
    return p;
}

static int
try_wait (void)
{
    int status;
    pid_t ret_pid;

    ret_pid = waitpid (pid, &status, WNOHANG);
    if (ret_pid == pid) {
	fprintf (stderr, "process (%d) exited with code %d\n",
		 pid, WEXITSTATUS (status));
	pid = 0;
	return WAIT_SUCCESS;
    } else {
	fprintf (stderr, "waitpid returns %d\n", ret_pid);
	return WAIT_FAILURE;
    }
}

static void
open_proc (pid_t pid)
{
    char buf [BUFSIZ];

    snprintf (buf, sizeof (buf), "/proc/%d/ctl",    pid);
    proc_fd.ctl = e_open (buf, O_WRONLY);

    snprintf (buf, sizeof (buf), "/proc/%d/status", pid);
    proc_fd.status = e_open (buf, O_RDONLY);

    snprintf (buf, sizeof (buf), "/proc/%d/as",     pid);
    proc_fd.as = e_open (buf, O_RDWR);
}

static void
close_proc (void)
{
    close (proc_fd.ctl);
    close (proc_fd.as);
    close (proc_fd.status);
}

static void
send_ctl_pcstop (long cmd)
{
    e_write (proc_fd.ctl, &cmd, sizeof (cmd));
}

static void
send_ctl_pcrun (long data)
{
    struct proc_ctl {
	long     cmd;
	long     data;  /* assuming no padding here */
    } ctl;

    ctl.cmd = PCRUN;
    ctl.data = data;
    e_write (proc_fd.ctl, &ctl, sizeof (ctl));
}

static void
send_ctl_pcsexit (int sysnum, ...)
{
    va_list ap;
    int i, sys;
    struct sys_ctl {
	long     cmd;
	sysset_t sysset;
    } ctl;

    premptyset (&ctl.sysset);
    va_start (ap, sysnum);
    for (i = 0; i < sysnum; i++) {
	sys = va_arg (ap, int);
	praddset (&ctl.sysset, sys);
    }
    va_end (ap);
    ctl.cmd = PCSEXIT;
    e_write (proc_fd.ctl, &ctl, sizeof (ctl));
}

static void
send_ctl_pcsfault (int fltnum, ...)
{
    va_list ap;
    int i, flt;
    struct flt_ctl {
	long     cmd;
	fltset_t fltset;
    } ctl;

    premptyset (&ctl.fltset);
    va_start (ap, fltnum);
    for (i = 0; i < fltnum; i++) {
	flt = va_arg (ap, int);
	praddset (&ctl.fltset, flt);
    }
    va_end (ap);
    ctl.cmd = PCSFAULT;
    e_write (proc_fd.ctl, &ctl, sizeof (ctl));
}

static void
get_pstatus (void)
{
    e_lseek (proc_fd.status, 0, SEEK_SET);
    e_read (proc_fd.status, &pstatus, sizeof (pstatus));
}

static void
init (struct command *cmd_p, int argc, char *argv [])
{
    struct sigaction act;
    long size;
    char *aout, *xml;

    assert (argc >= 2);
    aout = argv [1];

    cmd_p->line = NULL;

    cmd_name [CMD_NONE]  = "none";
    cmd_name [CMD_RUN]   = "run";
    cmd_name [CMD_QUIT]  = "quit";
    cmd_name [CMD_CONT]  = "cont";
    cmd_name [CMD_STEPI] = "stepi";
    cmd_name [CMD_STEP]  = "step";
    cmd_name [CMD_BREAK] = "break";
    cmd_name [CMD_PRINT] = "print";
    cmd_name [CMD_WHERE] = "where";
    cmd_name [CMD_WATCH] = "watch";

#if 0
    atexit (kill_child);
#endif

    act.sa_handler = sigint_handler;
    act.sa_flags = SA_RESTART;
    sigemptyset (&act.sa_mask);
    sigaction (SIGINT, &act, NULL);

    act.sa_handler = sigchld_handler;
    act.sa_flags = SA_RESTART;
    sigemptyset (&act.sa_mask);
    sigaction (SIGCHLD, &act, NULL);

    if (argc <= 2)
        return;

    /* XML init */
    xml = argv [2];
    if (!strcmp (".gz", xml + strlen (xml) -3)) {
	/* gzipped file */
	FILE *fp;
	char buf[BUFLEN], *gunzip_file;
	int  fd, size, ret;
	void *addr;
	struct stat statbuf;

	gunzip_file = tmpnam (NULL);
	snprintf (buf, BUFLEN, "gunzip -c %s > %s", xml, gunzip_file);
	ret = system (buf);
	if (ret == -1) {
	    perror ("system");
	    exit (1);
	}
	fp = fopen (gunzip_file, "r");
#if 1
	xml_doc  = xmlParseFile(gunzip_file);
	unlink (gunzip_file);
#else
	unlink (gunzip_file);
	fd = fileno (fp);
	fstat (fd, &statbuf);
	addr = mmap (0, statbuf.st_size, PROT_READ, MAP_SHARED, fd, 0);
	xml_doc = xmlParseMemory (addr, statbuf.st_size);
#endif
    } else {
	/* not gzipped file */
	xml_doc  = xmlParseFile(argv [2]);
    }
    xml_root = xmlDocGetRootElement (xml_doc);
    xmlXPathInit();
}

static void
read_command (struct command *cmd_p)
{
    char *line;
    assert (cmd_p != NULL);

    while (1) {
	line = readline ("[5m[7mx_debug> [m");
	if (line == NULL) {
	    do_quit (NULL);
	}
	if (line [0] != '\0') {
	    add_history (line);
	    break;
	}
	free (line);
    }

    if (cmd_p->line != NULL)
	free (cmd_p->line);
    cmd_p->line = line;
}

static void
parse_command (struct command *cmd_p)
{
    int i, ret;
    char *line, *p;

    line = cmd_p->line;

    for (i = 1; i < N_CMD; i++) {
	while (isspace(*line)) line++;
	if (!strncmp (line, cmd_name [i], strlen (cmd_name [i]))) {
	    cmd_p->type = (enum command_type) i;
	    break;
	}
    }
    if (cmd_name [i] == NULL) { /* not found */
	cmd_p->type = CMD_NONE;
    }

    switch (cmd_p->type) {
    case CMD_NONE:
    case CMD_RUN:
    case CMD_QUIT:
    case CMD_CONT:
    case CMD_STEPI:
    case CMD_WHERE:
    case CMD_STEP:
	break;
    case CMD_BREAK:
	line += strlen (cmd_name [CMD_BREAK]); /* skip "break" */
	while (isspace(*line)) line++;         /* skip whitespaces */
	if (!strncmp (line, "0x", 2)) {        /* address argument */
	    cmd_p->addr = (int *) strtol (line, NULL, 0);
    	} else if (isdigit(*line)) {           /* line-number argument */
	    cmd_p->addr = lineno2addr (atoi(line));
	} else {                               /* symbol argument */
	    cmd_p->addr = sym2addr (line);
	}
	break;
    case CMD_WATCH:
	line += strlen (cmd_name [CMD_WATCH]); /* skip "watch" */
	while (isspace(*line)) line++;         /* skip whitespaces */
	if (!strncmp (line, "0x", 2)) {        /* numeral argument */
	    cmd_p->addr = (int *) strtol (line, NULL, 0);
	} else {                               /* symbol argument */
	    char *p, *q;
	    p = q = strdup (line);
	    while (!isspace(*q)) q++;
	    *q = '\0';
	    cmd_p->addr = sym2addr (p);
	}
	while (!isspace(*line)) line++;
	while (isspace(*line))  line++;
	cmd_p->size = strtol (line, NULL, 0);
	break;
    case CMD_PRINT:
	line += strlen (cmd_name [CMD_PRINT]); /* skip "print" */
	while (isspace(*line)) line++;         /* skip whitespaces */
	if (line [0] == '/') {
	    cmd_p->fmt = line [1];
	    line += 2; while (isspace(*line)) line++; /* skip fmt */
	} else {
	    cmd_p->fmt = 'd';
	}
	if (!strncmp (line, "0x", 2)) {        /* numeral argument */
	    cmd_p->addr = (int *) strtol (line, NULL, 0);
	} else {                               /* symbol argument */
	    cmd_p->addr = sym2addr (line);
	}
	break;
    default:
	cmd_p->type = CMD_NONE;
    }
}

static void
dump_command (struct command *cmd_p)
{
    fprintf (stderr, "type = %s (%d)\n", cmd_name [cmd_p->type], cmd_p->type);
    if (cmd_p->line == NULL) {
	fprintf (stderr, "line = %p (null)\n", cmd_p->line);
    } else {
	fprintf (stderr, "line = %p (%s)\n", cmd_p->line, cmd_p->line);
    }
    fprintf (stderr, "arg  = %p\n", cmd_p->arg);
    fprintf (stderr, "addr = %p\n", cmd_p->addr);
    fprintf (stderr, "fmt  = %c\n", cmd_p->fmt);
    fprintf (stderr, "size = %d\n", cmd_p->size);
}

static void
do_run   (struct command *cmd_p, char *aout)
{
    kill_child ();
    pid = fork ();
    if (pid < 0) {
	perror ("fork"); exit (1);
    }

    if (pid == 0) { /* child process */
	pid = getpid ();
	setpgid (pid, pid); /* create new process group */
	open_proc (getpid ());
	send_ctl_pcsexit (1, SYS_execve);
	execve (aout, make_argv (aout, cmd_p->line), NULL);
	perror ("execve"); /* not reached if execve succeeded */
	exit (1);
    }

    /* parent process */
    open_proc (pid);
    /* FLTBOUNDS: segmentation fault, FLTACCESS: bus error */
    send_ctl_pcsfault (5, FLTBPT, FLTTRACE, FLTBOUNDS, FLTACCESS,
		       FLTWATCH);
    send_ctl_pcstop (PCWSTOP);
    print_child_state ();
    lineno2addr (5);
}

static void
do_quit  (struct command *cmd_p)
{
    kill_child ();
    exit (0);
}

static void
do_cont (struct command *cmd_p)
{
    if (pid == 0) {
	fprintf (stderr, "no process running\n");
	return;
    }
    adjust_if_breakpoint ();
    send_ctl_pcrun (PRCSIG | PRCFAULT);
    wait_child (1);
}


static void
do_stepi (struct command *cmd_p, int print_status)
{
    if (pid == 0) {
	fprintf (stderr, "no process running\n");
	return;
    }

    if (!adjust_if_breakpoint ()) {
	send_ctl_pcrun (PRCSIG | PRCFAULT | PRSTEP);
    }
    wait_child (print_status);
}

static void
do_step  (struct command *cmd_p)
{
    struct command command;
    void *pc = get_pc ();
    unsigned line;

    addr2lineno (pc);
    line = location.line;

    while (1) {
	do_stepi (&command, 0);
	addr2lineno (get_pc());
	if (line != location.line)
	    break;
    }
    print_child_state ();
}

static void
do_break (struct command *cmd_p)
{
    if (pid == 0) {
	fprintf (stderr, "no process running\n");
	return;
    }
    break_point.addr = cmd_p->addr;

    e_llseek (proc_fd.as, (offset_t)(unsigned) break_point.addr, SEEK_SET);
    e_read (proc_fd.as, &break_point.data, sizeof (break_point.data));

    e_llseek (proc_fd.as, (offset_t)(unsigned) break_point.addr, SEEK_SET);
    e_write (proc_fd.as, &trap_ta1, sizeof (trap_ta1));
}

static void
do_print (struct command *cmd_p)
{
    int int_data, nbytes;
    char char_data;
    char str_data [BUFSIZ];

    if (pid == 0) {
	fprintf (stderr, "no process running\n");
	return;
    }

    e_llseek (proc_fd.as, (offset_t)(unsigned) cmd_p->addr, SEEK_SET);

    switch (cmd_p->fmt) {
    case 'x':
	e_read (proc_fd.as, &int_data, sizeof (int_data));
	fprintf (stderr, "*(int*)(%x) => 0x%x\n", cmd_p->addr, int_data);
	break;
    case 'd':
	e_read (proc_fd.as, &int_data, sizeof (int_data));
	fprintf (stderr, "*(int*)(%x) => %d\n", cmd_p->addr, int_data);
	break;
    case 'c':
	e_read (proc_fd.as, &char_data, sizeof (int_data));
	fprintf (stderr, "*(char*)(%x) => '%c'\n", cmd_p->addr, int_data);
	break;
    case 's':
	/* some string may not be terminated with '\0'. 
	   so we limit the length of string to BUFSIZ. */
	nbytes = e_read (proc_fd.as, str_data, sizeof (str_data));
	str_data [nbytes - 1] = '\0';
	fprintf (stderr, "(char *)(%x) => '%s'\n", cmd_p->addr, str_data);
	break;
    default:
	assert (0);
	break;
    }
}

static void
do_where (struct command *cmd_p)
{
    struct rwindow rwindow;
    void *kwsa;
    int  *pc;

    if (pid == 0) {
	fprintf (stderr, "no process running\n");
	return;
    }

    pc = get_pc ();
    addr2lineno (pc);
    fprintf (stderr, "0x%x: ", pc);
    show_line ();

    for (kwsa = get_sp (); kwsa != 0; kwsa = (void *) rwindow.rw_in [6]) {
	if (!strcmp (location.func, "main"))
	    break;
	get_kwsa (kwsa, &rwindow);
	addr2lineno ((int *)(rwindow.rw_in [7])); /* return address */
	fprintf (stderr, "0x%x: ", rwindow.rw_in [7]);
	show_line ();
    }
}

static void
do_watch (struct command *cmd_p)
{
    struct watch_ctl {
	long      cmd;
	prwatch_t prwatch;
    } w;

    w.cmd = PCWATCH;
    w.prwatch.pr_wflags = ((cmd_p->size == 0) ? 0 : WA_WRITE);
    w.prwatch.pr_vaddr  = (uintptr_t) cmd_p->addr;
    w.prwatch.pr_size   = cmd_p->size;
    e_write (proc_fd.ctl, &w, sizeof (w));
}

static void
kill_child (void)
{
    if (pid != 0) {    /* child process is running */
	close_proc (); /* for sanity */
	kill (pid, SIGKILL);
	waitpid (pid, NULL, 0);
    }
}

static void
wait_child (int print_status)
{
    send_ctl_pcstop (PCWSTOP);

    /* in the case that child process has been dead, 'write' and 'read' fail.
       this implies the above 'send_ctl_pcstop' fails in the case.
       so, we need to check if the child is alive or not. */
    if (pid == 0)
	return;

    /* examine why child process stopped */
    if (print_status)
	print_child_state ();
}

static int
adjust_if_breakpoint (void)
{
    /* if we are in the breakpoint, we need to execute the
       overwritten instruction by single-stepping, and then
       re-plant the breakpoint. */
    if (get_pc () == break_point.addr) {
	/* write back the overwritten instruction to memory */
	e_llseek (proc_fd.as, (offset_t)(unsigned) break_point.addr, SEEK_SET);
	e_write (proc_fd.as, &break_point.data, sizeof (break_point.data));

	/* single stepping */
	send_ctl_pcrun (PRCSIG | PRCFAULT | PRSTEP);
	send_ctl_pcstop (PCWSTOP);

	/* re-plant the breakpoint */
	e_llseek (proc_fd.as, (offset_t)(unsigned) break_point.addr, SEEK_SET);
	e_write (proc_fd.as, &trap_ta1, sizeof (trap_ta1));

	return 1;
    }
    return 0;
}

static void *
get_pc (void)
{
    get_pstatus ();
    return (void *)pstatus.pr_lwp.pr_reg [R_PC];
}

static void *
get_sp (void)
{
    get_pstatus ();
    return (void *)pstatus.pr_lwp.pr_reg [R_O6]; /* %sp == %o6 */
}

static void *
get_fp (void)
{
    get_pstatus ();
    return (void *)pstatus.pr_lwp.pr_reg [R_I6]; /* %fp == %i6 */
}

/* kwsa - kernel window save area (= struct rwindow) */
static void
get_kwsa (void *kwsa, struct rwindow *rwindow) 
{
    e_llseek (proc_fd.as, (offset_t) (unsigned) kwsa, SEEK_SET);
    e_read  (proc_fd.as, rwindow, sizeof (struct rwindow));
}

static char **
make_argv (char *aout, char *line)
{
    int  i, argc, inword;
    char **argv;
    char *head, *p, *q;

    /* skip "run" and white spaces around */
    head = strdup (line);
    while (isspace(*head))  head++;
    while (!isspace(*head)) head++;
    while (isspace(*head))  head++;
    
    /* first count the elements */
    argc = 0; inword = 0;
    for (p = head; *p != '\0'; p++) {
	if (inword && isspace (*p)) {
	    inword = 0;
	} else if (!inword && !isspace (*p)) {
	    inword = 1;
	    argc++;
	}
    }
    argv = e_malloc (sizeof (char *) * (argc + 1)); /* 1 for NULL */

    /* then build argv */
    argv [0] = strdup (aout);
    p = q = head;
    for (i = 1; i <= argc; i++) {
	while (!(isspace (*q) || (*q == '\0'))) q++;
	*q = '\0';
	argv [i] = p;
	while (isspace (*q) || (*q == '\0')) q++;
	p = q;
    }

    argv [argc + 1] = NULL;
    return argv;
}

static void
sigint_handler (int signo)
{
    if (pid == 0)
	return;
    send_ctl_pcstop (PCDSTOP);
}

static void
sigchld_handler (int signo)
{
#if 0
    if (signo == SIGCHLD) {
	fprintf (stderr, "I received SIGCHLD\n");
    }
#endif
}

static void
print_child_state (void)
{
    int signo;
    void *addr, *pc;

    get_pstatus ();
    signo = pstatus.pr_lwp.pr_info.si_signo;
    pc    = (void *) pstatus.pr_lwp.pr_reg [R_PC];

    if (signo == SIGSEGV || signo == SIGILL || signo == SIGBUS) {
	addr  =  pstatus.pr_lwp.pr_info.si_addr;
	fprintf (stderr, "Stopped: SIG%s (addr=0x%x, pc=0x%x): ",
		 sig_name [signo], addr, pc);
    } else if (0 < signo && signo <= 37) {
	fprintf (stderr, "Stopped: SIG%s (pc=0x%x): ",
		 sig_name [signo], pc);
    } else {
	fprintf (stderr, "Stopped: pc=0x%x: ", pc);
    }

    addr2lineno (pc);
    show_line ();
}

static void
show_line (void)
{
    const char *file;

    if (location.file != NULL) {
	file = strrchr (location.file, '/');
	if (file == NULL) {
	    file = location.file;
	} else {
	    file++;
	}
	if (location.func == NULL) {
	    fprintf (stderr, "file = %s, line = %d\n", file, location.line);
	} else {
	    fprintf (stderr, "file = %s, func = %s, line = %d\n",
		     file, location.func, location.line);
	}
    }
}

static xmlNodeSetPtr
eval_xpath (char *xpath, xmlNodePtr from)
{
    xmlNodePtr node;
    xmlXPathObjectPtr  result;
    xmlXPathContextPtr context;

    context = xmlXPathNewContext(xml_doc);
    assert (context != NULL);
    context->node = from;
    result = xmlXPathEval (xpath, context);
    xmlXPathFreeContext (context);
#if DEBUG && 0
    xmlXPathDebugDumpObject(stdout, result, 0);
#endif
    return result->nodesetval;
}

static int *
sym2addr (char *sym)
{
    xmlNodeSetPtr result;
    void *pc = get_pc ();
    char buf [BUFLEN];
    int length, offset;
    unsigned long addr;
    char *addr_str;
#if 0
    struct timeval tv1, tv2;
    gettimeofday (&tv1, NULL);
#endif

    /* first, look for local variables */
    snprintf (buf, BUFLEN, 
	      "(//section[@name='.debug_info']//tag[@name='DW_TAG_subprogram' or @name='DW_TAG_lexical_block'][attribute[@name='DW_AT_high_pc']/@value>%lu and attribute[@name='DW_AT_low_pc']/@value<=%lu]//tag[@name='DW_TAG_variable'][attribute[@name='DW_AT_name'][@value='%s']])[last()]/attribute[@name='DW_AT_location']/description/text()",
	      pc, pc, sym);
    result = eval_xpath (buf, xml_root);
    length = xmlXPathNodeSetGetLength (result);
    if (length != 0) {
	assert (length == 1);
	addr_str = xmlXPathCastNodeSetToString (result);
	if (sscanf (addr_str, "DW_OP_fbreg: %d", &offset) == 1) {
	    printf ("offset = %d\n", offset);
	    return (int*)(get_fp () + offset);
	} else if (sscanf (addr_str, "DW_OP_addr: %u", &addr) == 1) {
	    printf ("addr = %u\n", addr);
	    return (int*)addr;
	} else {
	    fprintf (stderr, "sym2addr: unknown description(%s)\n", addr_str);
	    assert (0);
	}
    }

    /* then, look for global ones */
    snprintf (buf, BUFLEN, 
	      "(//section[@name='.debug_info']//tag[@name='DW_TAG_variable' and @level='1'][attribute[@name='DW_AT_name'][@value='%s']])/attribute[@name='DW_AT_location']/description/text()",
	      sym);
    result = eval_xpath (buf, xml_root);
    length = xmlXPathNodeSetGetLength (result);
    if (length >= 1) {
	assert (length == 1);
	addr_str = xmlXPathCastNodeSetToString (result);
	if (sscanf (addr_str, "DW_OP_addr: %u", &addr) == 1) {
	    printf ("addr = %u\n", addr);
	    return (int*)addr;
	} else {
	    fprintf (stderr, "sym2addr: unknown description(%s)\n", addr_str);
	    assert (0);
	}
    }

    /* finally, look for functions */
    snprintf (buf, BUFLEN, 
	      "(//section[@name='.debug_info']//tag[@name='DW_TAG_subprogram'][attribute[@name='DW_AT_name'][@value='%s']])/attribute[@name='DW_AT_low_pc']/@value",
	      sym);
    result = eval_xpath (buf, xml_root);
    length = xmlXPathNodeSetGetLength (result);
    if (length == 0) {
	fprintf (stderr, "sym2addr: no such symbol(%s)\n", sym);
	assert (0);
    }
    assert (length == 1);
    addr = xmlXPathCastNodeSetToNumber (result);
#if 0
    gettimeofday (&tv2, NULL);
    printf ("sym2addr: time %d, %d\n", tv2.tv_sec - tv1.tv_sec,
	    tv2.tv_usec - tv1.tv_usec);
#endif
    return (int*)addr;
}

static void
addr2lineno (int *addr)
{
    xmlNodeSetPtr result;
    xmlNodePtr    node;
    xmlChar*      attr_val;
    int length;
    int file_index, dir_index;
    char buf [BUFLEN];
#if 0
    struct timeval tv1, tv2;
    gettimeofday (&tv1, NULL);
#endif

    snprintf (buf, BUFLEN, 
	      "(//section[@name='.debug_line']//line[@address<=%lu])[last()]",
	      addr);
    result = eval_xpath (buf, xml_root);
    length = xmlXPathNodeSetGetLength (result);

    if (length == 0) {
	fprintf (stderr, "addr2lineno: no line for address(%lu)\n", addr);
	return;
    }

    assert (length == 1);
    node = result->nodeTab [0];
    attr_val = xmlGetProp (node, "line");
    location.line = atoi (attr_val);
    xmlFree (attr_val);

    attr_val = xmlGetProp (node, "file");
    file_index = atoi (attr_val);
    xmlFree (attr_val);

    snprintf (buf, BUFLEN,
	      "../../file_names/file[%lu]", file_index);
    result = eval_xpath (buf, node);
    node   = result->nodeTab [0];
    location.file = xmlXPathCastNodeSetToString (result);

    attr_val = xmlGetProp (node, "directory");
    dir_index = atoi (attr_val);
    xmlFree (attr_val);

    if (dir_index == 0) {
	location.dir = ".";
    } else {
	snprintf (buf, BUFLEN,
		  "../../include_directories/directory[%lu]", dir_index);
	result = eval_xpath (buf, node);
	location.dir = xmlXPathCastNodeSetToString (result);
    }

    snprintf (buf, BUFLEN,
	      "//section[@name='.debug_info']//tag[@name='DW_TAG_subprogram'][attribute[@name='DW_AT_high_pc']/@value>%lu and attribute[@name='DW_AT_low_pc']/@value<=%lu]/attribute[@name='DW_AT_name']/@value",

	      addr, addr);
    result = eval_xpath (buf, xml_root);
    location.func = xmlXPathCastNodeSetToString (result);
#if DEBUG && 0
    fprintf (stderr,
	     "addr2lineno: line = %d, func = %s, file = %s, dir = %s\n",
	    location.line, location.func, location.file, location.dir);
#endif
#if 0
    gettimeofday (&tv2, NULL);
    printf ("addr2lineno: time %d, %d\n", tv2.tv_sec - tv1.tv_sec,
	    tv2.tv_usec - tv1.tv_usec);
#endif
}

static int *
lineno2addr (int lineno)
{
    xmlNodeSetPtr result;
    char buf [BUFLEN];
    int file_index;
    unsigned int addr;
#if 0
    struct timeval tv1, tv2;
    gettimeofday (&tv1, NULL);
#endif

    addr2lineno (get_pc ()); /* to set location.file */
    snprintf (buf, BUFLEN, 
	      "//section[@name='.debug_line']//file[text()='%s']/@index",
	      location.file);
    result = eval_xpath (buf, xml_root);
    file_index = xmlXPathCastNodeSetToNumber (result);

    snprintf (buf, BUFLEN, 
	      "(//section[@name='.debug_line']//line[@line <= %lu"
	      "and @file='%d'])[last()]/@address", 
	      lineno, file_index);
    result = eval_xpath (buf, xml_root);
    addr = xmlXPathCastNodeSetToNumber (result);
#if 0
    gettimeofday (&tv2, NULL);
    printf ("lineno2addr: time %d, %d\n", tv2.tv_sec - tv1.tv_sec,
	    tv2.tv_usec - tv1.tv_usec);
#endif
    return (int *)addr;
}
/* ======================================================================== */
int
main (int argc, char *argv [])
{
    struct command command;

    if (argc < 2) {
	fprintf (stderr, "Usage: %s <a.out>\n", argv [0]);
	exit (1);
    }

    init (&command, argc, argv);

    while (1) { /* debugger main loop */
	read_command  (&command);
	parse_command (&command);
#if 0 /* for debugging */
	dump_command  (&command);
#endif

	switch (command.type) {
	case CMD_NONE:  /* do nothing */               break;
	case CMD_RUN:   do_run   (&command, argv [1]); break;
	case CMD_QUIT:  do_quit  (&command);           break;
	case CMD_CONT:  do_cont  (&command);           break;
	case CMD_STEPI: do_stepi (&command, 1);        break;
	case CMD_STEP:  do_step  (&command);           break;
	case CMD_BREAK: do_break (&command);           break;
	case CMD_PRINT: do_print (&command);           break;
	case CMD_WHERE: do_where (&command);           break;
        case CMD_WATCH: do_watch (&command);           break;
	default:        assert (0); exit (1);
	}
    }
}
/* ======================================================================== */
